from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional


class Engine(ABC):
    """This is an abstract class which is inherited for every engine. It acts as
    an interface to the engine and abstracts away things like running an analysis
    using the engine."""

    @abstractmethod
    def name(self) -> str:
        """Returns the name of the engine as used in the CLI."""
        raise NotImplementedError

    @abstractmethod
    def can_run_analysis(self, analysis_file: Path) -> bool:
        """Checks if this engine can run the analysis in the given file."""
        raise NotImplementedError

    @abstractmethod
    def run_local_analysis(
        self,
        analysis_file: Path,
        engine_path: Optional[str] = None,
        additional_cli_args: Optional[list[str]] = None,
    ):
        """Runs the analysis in the given file using the locally installed engine."""
        raise NotImplementedError

    @abstractmethod
    def run_containerized_analysis(
        self, analysis_file: Path, additional_cli_args: Optional[list[str]] = None
    ):
        """Runs the analysis in the given file in a container. This does not require the
        engine to be installed on the system."""
        raise NotImplementedError



from phylorun.engines.beast2 import BEAST2
from phylorun.engines.lphy import LPhy
from phylorun.engines.beastX import BEASTX
from phylorun.engines.engine import Engine
from phylorun.engines.revBayes import RevBayes

ENGINES: list[Engine] = [BEAST2(), BEASTX(), RevBayes(), LPhy()]



from pathlib import Path
import subprocess
from typing import Optional

from loguru import logger
from phylorun.engines.engine import Engine
from phylorun.utils.docker_utils import (
    create_image_if_needed,
    get_docker_client,
    run_and_print_command,
    start_container,
)


BINARY_URL = "https://github.com/revbayes/revbayes/releases/download/v1.3.1/revbayes-v1.3.1-linux64.tar.gz"


class RevBayes(Engine):
    def name(self) -> str:
        """Returns the name of the engine as used in the CLI."""
        return "revbayes"

    def can_run_analysis(self, analysis_file: Path) -> bool:
        """Checks if RevBayes can run the analysis in the given file."""
        if not analysis_file.name.endswith(".rev"):
            logger.debug("No Rev file: wrong file extension (.rev requried).")
            return False

        logger.debug("RevBayes file found.")
        return True

    def run_local_analysis(
        self,
        analysis_file: Path,
        engine_path: Optional[str] = None,
        additional_cli_args: Optional[list[str]] = None,
    ):
        """Runs the analysis in the given file using the locally installed engine."""
        engine_path = engine_path or "rb"
        if not engine_path:
            raise Exception("""No RevBayes binary found.
Use `phylorun --container your_analysis.rev` to use a docker container if you don't have RevBayes installed.
Use `phylorun --bin <path-to-binary> your_analysis.rev` to manually specify the RevBayes binary.
            """)

        additional_cli_args = additional_cli_args or []

        subprocess.run([engine_path, *additional_cli_args, analysis_file])

    def run_containerized_analysis(
        self, analysis_file: Path, additional_cli_args: Optional[list[str]] = None
    ):
        """Runs the analysis in the given file in a container. This does not require the
        engine to be installed on the system."""
        docker_client = get_docker_client()

        IMAGE_NAME = "revbayes:1.3.1"

        create_image_if_needed(
            docker_client,
            IMAGE_NAME,
            f"""FROM ubuntu:latest
            RUN apt-get update \\
                && apt-get install -y wget tar \\
                && wget {BINARY_URL} -O /revBayes.tgz \\
                && tar -xzf /revBayes.tgz -C /opt   \\
                && rm /revBayes.tgz
            """,
        )

        working_dir_is_data_dir = Path() == analysis_file.parent

        volumes = {str(analysis_file.parent.resolve()): {"bind": "/data", "mode": "rw"}}
        if not working_dir_is_data_dir:
            volumes[str(Path().resolve())] = {"bind": "/working", "mode": "rw"}

        container = start_container(
            docker_client,
            IMAGE_NAME,
            volumes=volumes,
        )

        try:
            if working_dir_is_data_dir:
                run_and_print_command(
                    container,
                    f"/opt/revbayes-v1.3.1/bin/rb {' '.join(additional_cli_args or [])} '/data/{analysis_file.name}'",
                    working_dir="/data",
                )
            else:
                run_and_print_command(
                    container,
                    f"/opt/revbayes-v1.3.1/bin/rb {' '.join(additional_cli_args or [])} '/data/{analysis_file.name}'",
                    working_dir="/working",
                )
        finally:
            container.stop()
            container.remove()



import os
from pathlib import Path
import subprocess
from typing import Optional
from phylorun.engines.engine import Engine
from xml.etree import ElementTree

from loguru import logger

from phylorun.utils.docker_utils import (
    create_image_if_needed,
    get_docker_client,
    run_and_print_command,
    start_container,
)


BINARY_URL = "https://github.com/beast-dev/beast-mcmc/releases/download/v10.5.0/BEAST_X_v10.5.0.tgz"
EB_URL = "https://raw.githubusercontent.com/easybuilders/easybuild-easyconfigs/refs/heads/main/easybuild/easyconfigs/b/beagle-lib/beagle-lib-4.0.1-GCC-12.3.0.eb"


class BEASTX(Engine):
    def name(self) -> str:
        """Returns the name of the engine as used in the CLI."""
        return "beastx"

    def can_run_analysis(self, analysis_file: Path) -> bool:
        """Checks if BEAST X can run the analysis in the given file."""
        try:
            xml = ElementTree.parse(analysis_file)
        except ElementTree.ParseError:
            logger.debug("No BEAST X file: no XML file.")
            return False

        root = xml.getroot()

        if root is None or root.tag.lower() != "beast":
            logger.debug("No BEAST X file: no root BEAST tag.")
            return False

        if not root.attrib.get("version", "").startswith("1.") and not root.attrib.get(
            "version", ""
        ).startswith("10."):
            logger.debug("No BEAST X file: wrong version (likely for BEAST 2).")
            return False

        if not any(child.tag.lower() == "mcmc" for child in root):
            logger.debug("No BEASTX file: no <mcmc> tag.")
            return False

        logger.debug("BEAST X file found.")

        return True

    def run_local_analysis(
        self,
        analysis_file: Path,
        engine_path: Optional[str] = None,
        additional_cli_args: Optional[list[str]] = None,
    ):
        """Runs the analysis in the given file using the locally installed engine."""
        engine_path = engine_path or self._find_binary_path()
        if not engine_path:
            raise Exception("""No BEAST X binary found.
Use `phylorun --container your_analysis.xml` to use a docker container if you don't have BEAST X installed.
Use `phylorun --bin <path-to-binary> your_analysis.xml` to manually specify the BEAST X binary.
            """)

        additional_cli_args = additional_cli_args or []

        subprocess.run([engine_path, *additional_cli_args, analysis_file])

    def _find_binary_path(self) -> Optional[str]:
        if path := os.environ.get("BEAST"):
            return path

        possible_paths = (
            list(Path("/Applications").glob("BEAST X*/bin/beast"))
            + list(Path("/Applications").glob("BEAST X*/bin/beast"))
            + list(Path("~").glob("beast*/bin/beast"))
        )
        possible_paths = sorted(possible_paths)

        if not possible_paths:
            return None

        return str(possible_paths[-1])

    def run_containerized_analysis(
        self, analysis_file: Path, additional_cli_args: Optional[list[str]] = None
    ):
        """Runs the analysis in the given file in a container. This does not require the
        engine to be installed on the system."""
        docker_client = get_docker_client()

        IMAGE_NAME = "beastx:10.5.0"

        create_image_if_needed(
            docker_client,
            IMAGE_NAME,
            f"""FROM ubuntu:latest
            RUN apt-get update \\
                && apt-get install -y wget tar openjdk-17-jdk python3-venv python3-pip \\
                && wget {BINARY_URL} -O /BEAST.tgz \\
                && tar -xzf /BEAST.tgz -C /opt   \\
                && rm /BEAST.tgz
            """,
        )

        working_dir_is_data_dir = Path() == analysis_file.parent

        volumes = {str(analysis_file.parent.resolve()): {"bind": "/data", "mode": "rw"}}
        if not working_dir_is_data_dir:
            volumes[str(Path().resolve())] = {"bind": "/working", "mode": "rw"}

        container = start_container(
            docker_client,
            IMAGE_NAME,
            volumes=volumes,
        )

        try:
            if working_dir_is_data_dir:
                run_and_print_command(
                    container,
                    f"/opt/BEASTv10.5.0/bin/beast -java {' '.join(additional_cli_args or [])} '/data/{analysis_file.name}'",
                    working_dir="/data",
                )
            else:
                run_and_print_command(
                    container,
                    f"/opt/BEASTv10.5.0/bin/beast -java {' '.join(additional_cli_args or [])} '/data/{analysis_file.name}'",
                    working_dir="/working",
                )
        finally:
            container.stop()
            container.remove()



from pathlib import Path

from phylorun.engines.beast2 import BEAST2


def to_file(text: str, path: str):
    path.write_text(text)
    return path


def test_simple_beast_config_is_ok(tmp_path: Path):
    path = to_file(
        """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
        <beast namespace="beast.core" required="BEAST.base v2.7.7" version="2.7">
            <data></data>
            <run></run>
        </beast>""",
        tmp_path / "analysis.xml",
    )
    assert BEAST2().can_run_analysis(path)


def test_simple_beast_config_with_alignment_is_ok(tmp_path):
    path = to_file(
        """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
        <beast namespace="beast.core" required="BEAST.base v2.7.7" version="2.7">
            <alignment></alignment>
            <run></run>
        </beast>""",
        tmp_path / "analysis.xml",
    )
    assert BEAST2().can_run_analysis(path)


def test_beast1_xml_fails(tmp_path):
    path = to_file(
        """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
        <beast version="1.10.4">
            <data></data>
            <run></run>
        </beast>""",
        tmp_path / "analysis.xml",
    )
    assert not BEAST2().can_run_analysis(path)


def test_missing_data_tag_fails(tmp_path):
    path = to_file(
        """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
        <beast namespace="beast.core" required="BEAST.base v2.7.7" version="2.7">
            <run></run>
        </beast>""",
        tmp_path / "analysis.xml",
    )
    assert not BEAST2().can_run_analysis(path)


def test_missing_run_tag_fails(tmp_path):
    path = to_file(
        """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
        <beast namespace="beast.core" required="BEAST.base v2.7.7" version="2.7">
            <data></data>
        </beast>""",
        tmp_path / "analysis.xml",
    )
    assert not BEAST2().can_run_analysis(path)


def test_empty_file_fails(tmp_path):
    path = to_file(
        "",
        tmp_path / "analysis.xml",
    )
    assert not BEAST2().can_run_analysis(path)


def test_non_xml_file_fails(tmp_path):
    path = to_file(
        """
        config:
            - this is not a XML file
        """,
        tmp_path / "analysis.xml",
    )
    assert not BEAST2().can_run_analysis(path)


def test_missing_beast_tag_fails(tmp_path):
    path = to_file(
        """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
        <some-other-tag></some-other-tag>""",
        tmp_path / "analysis.xml",
    )
    assert not BEAST2().can_run_analysis(path)



from pathlib import Path

from phylorun.engines.revBayes import RevBayes


def to_file(text: str, path: str):
    path.write_text(text)
    return path


def test_rev_file_is_ok(tmp_path: Path):
    path = to_file(
        "",
        tmp_path / "analysis.rev",
    )
    assert RevBayes().can_run_analysis(path)


def test_non_rev_file_is_ok(tmp_path: Path):
    path = to_file(
        "",
        tmp_path / "analysis.txt",
    )
    assert not RevBayes().can_run_analysis(path)



from pathlib import Path
from typing import Optional

from phylorun.engines.beast2 import BEAST2
from phylorun.engines.engine import Engine

from loguru import logger
import subprocess

import os


class LPhy(Engine):
    def name(self) -> str:
        """Returns the name of the engine as used in the CLI."""
        return "lphy"

    def can_run_analysis(self, analysis_file: Path) -> bool:
        """Checks if LPhy engine can run the analysis in the given file."""
        if not analysis_file.name.endswith(".lphy"):
            logger.debug("No LPhy file: wrong file extension (.lphy requried).")
            return False

        logger.debug("LPhy file found.")
        return True

    def run_local_analysis(
        self,
        analysis_file: Path,
        engine_path: Optional[str] = None,
        additional_cli_args: Optional[list[str]] = None,
    ):
        """Runs the analysis in the given file using the locally installed engine."""
        engine_path = engine_path or self._find_binary_path()
        if not engine_path:
            raise Exception("""No lphybeast binary found.
Use `phylorun --bin <path-to-binary> your_analysis.xml` to manually specify the lphybeast binary.
            """)

        additional_lphy_cli_args = [
            arg for arg in additional_cli_args if not arg.startswith("--beast2")
        ]
        additional_beast_cli_args = [
            arg.removeprefix("--beast2")
            for arg in additional_cli_args
            if arg.startswith("--beast2")
        ] or ["-working"]

        env = os.environ.copy()
        if beast_path := self._find_beast_path():
            env["BEAST"] = beast_path

        # run lphybeast
        subprocess.run(
            ["sh", engine_path, *additional_lphy_cli_args, analysis_file], env=env
        )

        # run BEAST 2
        beast2_file = analysis_file.parent / (analysis_file.stem + ".xml")
        beast2 = BEAST2()
        beast2.run_local_analysis(
            beast2_file, additional_cli_args=additional_beast_cli_args
        )

    def _find_binary_path(self) -> Optional[str]:
        if path := os.environ.get("BEAST"):
            return path

        possible_paths = list(
            Path("/Users").glob(
                "*/Library/Application Support/BEAST/2.*/lphybeast/bin/lphybeast"
            )
        )
        possible_paths = sorted(possible_paths)

        if not possible_paths:
            return None

        return str(possible_paths[-1])

    def _find_beast_path(self) -> Optional[str]:
        if path := os.environ.get("BEAST"):
            return path

        possible_paths = list(Path("/Applications").glob("BEAST 2.*"))
        possible_paths = sorted(possible_paths)

        if not possible_paths:
            return None

        return str(possible_paths[-1])

    def run_containerized_analysis(
        self, analysis_file: Path, additional_cli_args: Optional[list[str]] = None
    ):
        """Runs the analysis in the given file in a container. This does not require the
        engine to be installed on the system."""
        raise NotImplementedError



from pathlib import Path
from typing import Optional

import click

from phylorun.engines import ENGINES
from phylorun.engines.engine import Engine


CONTEXT_SETTINGS = dict(ignore_unknown_options=True, allow_extra_args=True)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    "--engine",
    type=click.Choice(["beastx", "beast2", "revbayes"], case_sensitive=False),
    required=False,
    help="Select engine explicitly: beastx | beast2 | revbayes.",
)
@click.option(
    "--bin",
    "engine_path",
    type=click.Path(exists=True, dir_okay=False, path_type=str),
    required=False,
    help="Path to the engine binary to use for local runs.",
)
@click.option(
    "--container",
    is_flag=True,
    help="Run inside a containerized environment (no local engine install required).",
)
@click.argument(
    "analysis_file",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
)
@click.pass_context
def cli(
    ctx: click.Context,
    engine: Optional[str],
    engine_path: Optional[str],
    container: bool,
    analysis_file: Path,
) -> None:
    """Run BEAST X, BEAST 2, or RevBayes analyses from a single CLI.

    Examples:
      phylorun someModel.xml
      phylorun --beast2 someModel.xml
      phylorun --bin /path/to/beast someModel.xml
      phylorun --container someModel.rev
    """

    # Choose engine: flag forces selection; otherwise auto-detect
    selected_engine: Engine | None = None

    if engine:
        for e in ENGINES:
            if e.name() == engine:
                selected_engine = e
                break

        if selected_engine is None:
            raise click.ClickException(f"Engine '{engine}' is not available.")
    else:
        for potentialEngine in ENGINES:
            if potentialEngine.can_run_analysis(analysis_file):
                selected_engine = potentialEngine
                break

        if selected_engine is None:
            raise click.ClickException(
                f"Could not detect a supported engine for file '{analysis_file}'."
            )

    additional_args = list(ctx.args) if ctx.args else None

    if container:
        selected_engine.run_containerized_analysis(analysis_file, additional_args)
    else:
        selected_engine.run_local_analysis(analysis_file, engine_path, additional_args)


if __name__ == "__main__":
    cli()



import io
import sys
from typing import Optional
import docker
from docker.models.containers import Container
from docker.errors import DockerException, ImageNotFound
from loguru import logger


def get_docker_client() -> docker.DockerClient:
    """Instantiate and return a Docker client."""
    try:
        return docker.from_env()
    except DockerException:
        logger.error("Docker could not be instantiated. Is it installed and running?")
        raise Exception(
            "Docker could not be instantiated. Is it installed and running?"
        )


def create_image_if_needed(
    client: docker.DockerClient, image_name: str, docker_file: str
):
    """Ensure a Docker image exists, building it from a Dockerfile string if necessary."""
    try:
        client.images.get(image_name)
        # image already created
        return
    except ImageNotFound:
        # we still need to create the image
        ...

    logger.info(
        "Setting up docker container. This might take a while, but only has to be done once."
    )

    docker_file_bytes = io.BytesIO(docker_file.encode("utf-8"))

    client.images.build(
        fileobj=docker_file_bytes,
        tag=image_name,
        rm=True,
        platform="linux/x86_64",
    )


def start_container(
    client: docker.DockerClient, image_name: str, **kwargs
) -> Container:
    """Start a Docker container from a specified image.

    Args:
        client (docker.DockerClient): The Docker client.
        image_name (str): Name of the Docker image.
        **kwargs: Additional keyword arguments passed to container run.

    Returns:
        Container: The started Docker container.
    """
    return client.containers.run(
        image_name, "sleep infinity", detach=True, platform="linux/x86_64", **kwargs
    )


def run_and_print_command(
    container: Container,
    command: str,
    user: Optional[str] = None,
    working_dir: Optional[str] = None,
):
    """Run a command inside a Docker container and print its output to stdout and stderr.

    Args:
        container (Container): The Docker container where the command will be executed.
        command (str): The command to execute.
        user (Optional[str]): User to run the command as (if specified).
        working_dir (Optional[str]): The working directory inside the container (if specified).
    """
    result = container.exec_run(
        command, stream=True, demux=True, user=user or "root", workdir=working_dir
    )

    for stdout, stderr in result.output:
        if stdout:
            print(stdout.decode(), end="")
        if stderr:
            print(stderr.decode(), end="", file=sys.stderr)
